/*
Copyright © 2014 Andreas Enge <andreas.enge@inria.fr>

This file is part of pari-gnump.

Pari-gnump is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

Pari-gnump is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Pari-gnump.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "pari-gnump.h"

/****************************************************************************/
/*                                                                          */
/* Test functions                                                           */
/*                                                                          */
/****************************************************************************/

int test_z_to_GEN (mpz_srcptr z)
{
   int ret;
   mpz_t z2;
   GEN x;

   mpz_init (z2);

   x = mpz_get_GEN (z);
   mpz_set_GEN (z2, x);
   ret = mpz_cmp (z, z2);

   mpz_clear (z2);

   return (ret);
}

/****************************************************************************/

int test_GEN_to_z (GEN x)
{
   int ret;
   mpz_t z;
   GEN x2;

   mpz_init (z);

   mpz_set_GEN (z, x);
   x2 = mpz_get_GEN (z);

   mpz_clear (z);

   return (cmpii (x, x2));
}

/****************************************************************************/

int test_q_to_GEN (mpq_srcptr q)
{
   int ret;
   mpq_t q2;
   GEN x;

   mpq_init (q2);

   x = mpq_get_GEN (q);
   mpq_set_GEN (q2, x);
   ret = mpq_cmp (q, q2);

   mpq_clear (q2);

   return (ret);
}

/****************************************************************************/

int test_GEN_to_q (GEN x)
{
   int ret;
   mpq_t q;
   GEN x2;

   mpq_init (q);

   mpq_set_GEN (q, x);
   x2 = mpq_get_GEN (q);

   mpq_clear (q);

   return (gcmp (x, x2));
}

/****************************************************************************/

int test_f_to_GEN (mpfr_srcptr f)
{
   int ret;
   mpfr_t f2;
   GEN x;

   mpfr_init2 (f2, mpfr_get_prec (f));

   x = mpfr_get_GEN (f);
   mpfr_set_GEN (f2, x, MPFR_RNDN);
   ret = mpfr_cmp (f, f2);

   mpfr_clear (f2);

   return (ret);
}

/****************************************************************************/

int test_f_to_GEN_stack (mpfr_srcptr f)
{
   int ret;
   mpfr_t f2;
   GEN x;

   pari_mpfr_init2 (f2, mpfr_get_prec (f));

   x = mpfr_get_GEN (f);
   mpfr_set_GEN (f2, x, MPFR_RNDN);
   ret = mpfr_cmp (f, f2);

   return (ret);
}

/****************************************************************************/

int test_GEN_to_f (GEN x)
{
   int ret;
   mpfr_t f;
   GEN x2;

   mpfr_init (f);

   mpfr_set_GEN (f, x, MPFR_RNDN);
   x2 = mpfr_get_GEN (f);

   mpfr_clear (f);

   return (gcmp (x, x2));
}

/****************************************************************************/

int test_GEN_to_f_stack (GEN x)
{
   int ret;
   mpfr_t f;
   GEN x2;

   pari_mpfr_init_set_GEN (f, x, 53);
   x2 = mpfr_get_GEN (f);

   return (gcmp (x, x2));
}

/****************************************************************************/

int test_c_to_GEN (mpc_srcptr c)
{
   int ret;
   mpc_t c2;
   GEN x;

   mpc_init3 (c2, mpfr_get_prec (mpc_realref (c)),
                  mpfr_get_prec (mpc_imagref (c)));

   x = mpc_get_GEN (c);
   mpc_set_GEN (c2, x, MPC_RNDNN);
   ret = mpc_cmp (c, c2);

   mpc_clear (c2);

   return (ret);
}

/****************************************************************************/

int test_c_to_GEN_stack (mpc_srcptr c)
{
   int ret;
   mpc_t c2;
   GEN x;

   pari_mpc_init3 (c2, mpfr_get_prec (mpc_realref (c)),
                       mpfr_get_prec (mpc_imagref (c)));

   x = mpc_get_GEN (c);
   mpc_set_GEN (c2, x, MPC_RNDNN);
   ret = mpc_cmp (c, c2);

   return (ret);
}

/****************************************************************************/

int test_GEN_to_c (GEN x)
{
   int ret;
   mpc_t c;
   GEN x2;

   mpc_init2 (c, 100);

   mpc_set_GEN (c, x, MPC_RNDNN);
   x2 = mpc_get_GEN (c);

   mpc_clear (c);

   return (gequal (x, x2) != 1);
}

/****************************************************************************/

int test_GEN_to_c_stack (GEN x)
{
   int ret;
   mpc_t c;
   GEN x2;

   pari_mpc_init_set_GEN (c, x, 53);
   x2 = mpc_get_GEN (c);

   return (gequal (x, x2) != 1);
}

/****************************************************************************/

void test ()
{
   mpz_t z, z2;
   mpq_t q, q2;
   mpfr_t f, f2;
   mpc_t c, c2;
   gmp_randstate_t rand;
   GEN x;
   mpfr_prec_t prec;

   pari_init (1000000, 0);
   gmp_randinit_default (rand);

   mpz_init (z);
   mpz_set_si (z, 3);
   mpz_mul_2exp (z, z, 200);
   assert (test_z_to_GEN (z) == 0);
   mpz_neg (z, z);
   assert (test_z_to_GEN (z) == 0);
   mpz_set_si (z, 0);
   assert (test_z_to_GEN (z) == 0);
   mpz_clear (z);

   x = shifti (stoi (3), 200);
   assert (test_GEN_to_z (x) == 0);
   x = negi (x);
   assert (test_GEN_to_z (x) == 0);
   assert (test_GEN_to_z (gen_0) == 0);

   mpq_init (q);
   mpq_set_si (q, 3, 1);
   mpq_div_2exp (q, q, 200);
   assert (test_q_to_GEN (q) == 0);
   mpq_neg (q, q);
   assert (test_q_to_GEN (q) == 0);
   mpq_set_ui (q, 3, 1);
   assert (test_q_to_GEN (q) == 0);
   mpq_set_ui (q, 0, 1);
   assert (test_q_to_GEN (q) == 0);
   mpq_clear (q);

   x = gmul2n (stoi (3), -200);
   assert (test_GEN_to_q (x) == 0);
   x = gneg (x);
   assert (test_GEN_to_q (x) == 0);
   assert (test_GEN_to_q (stoi (3)) == 0);
   assert (test_GEN_to_q (gen_0) == 0);

   mpfr_init (f);
   mpfr_set_si (f, 3, MPFR_RNDN);
   mpfr_div_2ui (f, f, 200, MPFR_RNDN);
   assert (test_f_to_GEN (f) == 0);
   assert (test_f_to_GEN_stack (f) == 0);
   mpfr_neg (f, f, MPFR_RNDN);
   assert (test_f_to_GEN (f) == 0);
   assert (test_f_to_GEN_stack (f) == 0);
   mpfr_set_ui (f, 0, MPFR_RNDN);
   assert (test_f_to_GEN (f) == 0);
   assert (test_f_to_GEN_stack (f) == 0);
   for (prec = 2; prec < 200; prec++) {
      mpfr_set_prec (f, prec);
      mpfr_urandomb (f, rand);
      assert (test_f_to_GEN (f) == 0);
      assert (test_f_to_GEN_stack (f) == 0);
   }

   mpfr_clear (f);

   x = gmul2n (stor (3, 6), -200);
   assert (test_GEN_to_f (x) == 0);
   x = gneg (x);
   assert (test_GEN_to_f (x) == 0);
   assert (test_GEN_to_f (stor (3, 6)) == 0);
   assert (test_GEN_to_f (real_0_bit (-230)) == 0);
   assert (test_GEN_to_f (real_0_bit (230)) == 0);
   assert (test_GEN_to_f (gen_0) == 0);

   x = gmul2n (stor (3, 6), -200);
   assert (test_GEN_to_f_stack (x) == 0);
   x = gneg (x);
   assert (test_GEN_to_f_stack (x) == 0);
   assert (test_GEN_to_f_stack (stor (3, 6)) == 0);
   assert (test_GEN_to_f_stack (real_0_bit (-230)) == 0);
   assert (test_GEN_to_f_stack (real_0_bit (230)) == 0);

   mpc_init2 (c, 123);
   mpc_set_ui_ui (c, 3, 5, MPC_RNDNN);
   mpc_div_2ui (c, c, 100, MPC_RNDNN);
   assert (test_c_to_GEN (c) == 0);
   assert (test_c_to_GEN_stack (c) == 0);
   mpc_set_ui_ui (c, 0, 5, MPC_RNDNN);
   assert (test_c_to_GEN (c) == 0);
   assert (test_c_to_GEN_stack (c) == 0);
   mpc_set_ui_ui (c, 3, 0, MPC_RNDNN);
   assert (test_c_to_GEN (c) == 0);
   assert (test_c_to_GEN_stack (c) == 0);
   for (prec = 2; prec < 200; prec++) {
      mpc_set_prec (c, prec);
      mpc_urandom (c, rand);
      assert (test_c_to_GEN (c) == 0);
      assert (test_c_to_GEN_stack (c) == 0);
   }

   x = gmul2n (mkcomplex (stor (3, 6), stor (5, 6)), -200);
   assert (test_GEN_to_c (x) == 0);
   assert (test_GEN_to_c_stack (x) == 0);
   x = gmul2n (mkcomplex (gen_0, stor (5, 6)), -200);
   assert (test_GEN_to_c (x) == 0);
   assert (test_GEN_to_c_stack (x) == 0);
   x = gmul2n (mkcomplex (stor (3, 6), gen_0), -200);
   assert (test_GEN_to_c (x) == 0);
   assert (test_GEN_to_c_stack (x) == 0);
   
   mpc_clear (c);
   
   pari_close ();
   gmp_randclear (rand);
}

/****************************************************************************/

int main ()
{
   test ();
   printf ("All library tests passed.\n");

   return 0;
}
